@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 #
 # Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
-# Copyright (c) 2008-2014 California Institute of Technology.
+# Copyright (c) 2008-2015 California Institute of Technology.
 # License: 3-clause BSD.  The full license text is available at:
 #  - http://trac.mystic.cacr.caltech.edu/project/pathos/browser/dill/LICENSE
 
@@ -24,10 +24,14 @@ __license__ = """
 """ + __license__
 
 from .dill import dump, dumps, load, loads, dump_session, load_session, \
-    Pickler, Unpickler, register, copy, pickle, pickles, HIGHEST_PROTOCOL, \
-    DEFAULT_PROTOCOL, PicklingError, UnpicklingError
+    Pickler, Unpickler, register, copy, pickle, pickles, \
+    HIGHEST_PROTOCOL, DEFAULT_PROTOCOL, PicklingError, UnpicklingError, \
+    HANDLE_FMODE, CONTENTS_FMODE, FILE_FMODE
 from . import source, temp, detect
 
+# get global settings
+from .settings import settings
+
 # make sure "trace" is turned off
 detect.trace(False)
 
@@ -73,11 +77,14 @@ def load_types(pickleable=True, unpickleable=True):
     # add corresponding types from objects to types
     reload(types)
 
-def __extend():
-    from .dill import _extend
-    _extend()
+def extend(use_dill=True):
+    '''add (or remove) dill types to/from pickle'''
+    from .dill import _revert_extension, _extend
+    if use_dill: _extend()
+    else: _revert_extension()
     return
-__extend(); del __extend
+
+extend()
 
 def license():
     """print license"""
