@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 #
 # Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
-# Copyright (c) 2008-2014 California Institute of Technology.
+# Copyright (c) 2008-2015 California Institute of Technology.
 # License: 3-clause BSD.  The full license text is available at:
 #  - http://trac.mystic.cacr.caltech.edu/project/pathos/browser/dill/LICENSE
 """
@@ -14,10 +14,10 @@ Initial port to python3 by Jonathan Dobson, continued by mmckerns.
 Test against "all" python types (Std. Lib. CH 1-15 @ 2.7) by mmckerns.
 Test against CH16+ Std. Lib. ... TBD.
 """
-__all__ = ['dump','dumps','load','loads','dump_session','load_session',\
-           'Pickler','Unpickler','register','copy','pickle','pickles',\
-           'HIGHEST_PROTOCOL','DEFAULT_PROTOCOL',\
-           'PicklingError','UnpicklingError']
+__all__ = ['dump','dumps','load','loads','dump_session','load_session',
+           'Pickler','Unpickler','register','copy','pickle','pickles',
+           'HIGHEST_PROTOCOL','DEFAULT_PROTOCOL','PicklingError',
+           'UnpicklingError','HANDLE_FMODE','CONTENTS_FMODE','FILE_FMODE']
 
 import logging
 log = logging.getLogger("dill")
@@ -30,7 +30,9 @@ def _trace(boolean):
 
 import os
 import sys
-PY3 = (hex(sys.hexversion) >= '0x30000f0')
+diff = None
+_use_diff = False
+PY3 = (sys.hexversion >= 0x30000f0)
 if PY3: #XXX: get types from dill.objtypes ?
     import builtins as __builtin__
     from pickle import _Pickler as StockPickler, _Unpickler as StockUnpickler
@@ -68,17 +70,35 @@ from weakref import ReferenceType, ProxyType, CallableProxyType
 from functools import partial
 from operator import itemgetter, attrgetter
 # new in python2.5
-if hex(sys.hexversion) >= '0x20500f0':
+if sys.hexversion >= 0x20500f0:
     from types import MemberDescriptorType, GetSetDescriptorType
+# new in python3.3
+if sys.hexversion < 0x03030000:
+    FileNotFoundError = IOError
 try:
     import ctypes
     HAS_CTYPES = True
+    IS_PYPY = True
 except ImportError:
     HAS_CTYPES = False
+    IS_PYPY = False
 try:
     from numpy import ufunc as NumpyUfuncType
+    from numpy import ndarray as NumpyArrayType
+    def ndarraysubclassinstance(obj):
+        try: # check if is ndarray, and elif is subclass of ndarray
+            if getattr(obj, '__class__', type) is NumpyArrayType: return False
+            elif not isinstance(obj, NumpyArrayType): return False
+        except ReferenceError: return False # handle 'R3' weakref in 3.x
+        # verify that __reduce__ has not been overridden
+        NumpyInstance = NumpyArrayType((0,),'int8')
+        if id(obj.__reduce_ex__) == id(NumpyInstance.__reduce_ex__) and \
+           id(obj.__reduce__) == id(NumpyInstance.__reduce__): return True
+        return False
 except ImportError:
     NumpyUfuncType = None
+    NumpyArrayType = None
+    def ndarraysubclassinstance(obj): return False
 
 # make sure to add these 'hand-built' types to _typemap
 if PY3:
@@ -118,38 +138,67 @@ try:
     ExitType = None     # IPython.core.autocall.ExitAutocall
     singletontypes = ['exit', 'quit', 'get_ipython']
 except NameError:
-    ExitType = type(exit)
+    try: ExitType = type(exit) # apparently 'exit' can be removed
+    except NameError: ExitType = None
     singletontypes = []
 
+### File modes
+# Pickles the file handle, preserving mode. The position of the unpickled
+# object is as for a new file handle.
+HANDLE_FMODE = 0
+# Pickles the file contents, creating a new file if on load the file does
+# not exist. The position = min(pickled position, EOF) and mode is chosen
+# as such that "best" preserves behavior of the original file.
+CONTENTS_FMODE = 1
+# Pickles the entire file (handle and contents), preserving mode and position.
+FILE_FMODE = 2
+
 ### Shorthands (modified from python2.5/lib/pickle.py)
 def copy(obj, *args, **kwds):
     """use pickling to 'copy' an object"""
     return loads(dumps(obj, *args, **kwds))
 
-def dump(obj, file, protocol=None, byref=False):
+def dump(obj, file, protocol=None, byref=None, fmode=None, recurse=None):#, strictio=None):
     """pickle an object to a file"""
-    if protocol is None: protocol = DEFAULT_PROTOCOL
+    from .settings import settings
+    strictio = False #FIXME: strict=True needs cleanup
+    if protocol is None: protocol = settings['protocol']
+    if byref is None: byref = settings['byref']
+    if fmode is None: fmode = settings['fmode']
+    if recurse is None: recurse = settings['recurse']
     pik = Pickler(file, protocol)
-    pik._main_module = _main_module
-    _byref = pik._byref
+    pik._main = _main_module
+    # apply kwd settings
     pik._byref = bool(byref)
+    pik._strictio = bool(strictio)
+    pik._fmode = fmode
+    pik._recurse = bool(recurse)
+    # hack to catch subclassed numpy array instances
+    if NumpyArrayType and ndarraysubclassinstance(obj):
+        @register(type(obj))
+        def save_numpy_array(pickler, obj):
+            log.info("Nu: (%s, %s)" % (obj.shape,obj.dtype))
+            npdict = getattr(obj, '__dict__', None)
+            f, args, state = obj.__reduce__()
+            pik.save_reduce(_create_array, (f, args, state, npdict), obj=obj)
+            return
+    # end hack
     pik.dump(obj)
-    pik._byref = _byref
     return
 
-def dumps(obj, protocol=None, byref=False):
+def dumps(obj, protocol=None, byref=None, fmode=None, recurse=None):#, strictio=None):
     """pickle an object to a string"""
     file = StringIO()
-    dump(obj, file, protocol, byref)
+    dump(obj, file, protocol, byref, fmode, recurse)#, strictio)
     return file.getvalue()
 
 def load(file):
     """unpickle an object from a file"""
     pik = Unpickler(file)
-    pik._main_module = _main_module
+    pik._main = _main_module
     obj = pik.load()
     if type(obj).__module__ == _main_module.__name__: # point obj class to main
-        try: obj.__class__ == getattr(pik._main_module, type(obj).__name__)
+        try: obj.__class__ == getattr(pik._main, type(obj).__name__)
         except AttributeError: pass # defined in a file
    #_main_module.__dict__.update(obj.__dict__) #XXX: should update globals ?
     return obj
@@ -170,65 +219,143 @@ def loads(str):
 ### End: Shorthands ###
 
 ### Pickle the Interpreter Session
-def dump_session(filename='/tmp/session.pkl', main_module=_main_module):
+def _module_map():
+    """get map of imported modules"""
+    from collections import defaultdict
+    modmap = defaultdict(list)
+    items = 'items' if PY3 else 'iteritems'
+    for name, module in getattr(sys.modules, items)():
+        if module is None:
+            continue
+        for objname, obj in module.__dict__.items():
+            modmap[objname].append((obj, name))
+    return modmap
+
+def _lookup_module(modmap, name, obj, main_module): #FIXME: needs work
+    """lookup name if module is imported"""
+    for modobj, modname in modmap[name]:
+        if modobj is obj and modname != main_module.__name__:
+            return modname
+
+def _stash_modules(main_module):
+    modmap = _module_map()
+    imported = []
+    original = {}
+    items = 'items' if PY3 else 'iteritems'
+    for name, obj in getattr(main_module.__dict__, items)():
+        source_module = _lookup_module(modmap, name, obj, main_module)
+        if source_module:
+            imported.append((source_module, name))
+        else:
+            original[name] = obj
+    if len(imported):
+        import types
+        newmod = types.ModuleType(main_module.__name__)
+        newmod.__dict__.update(original)
+        newmod.__dill_imported = imported
+        return newmod
+    else:
+        return original
+
+def _restore_modules(main_module):
+    if '__dill_imported' not in main_module.__dict__:
+        return
+    imports = main_module.__dict__.pop('__dill_imported')
+    for module, name in imports:
+        exec("from %s import %s" % (module, name), main_module.__dict__)
+
+#NOTE: 06/03/15 renamed main_module to main
+def dump_session(filename='/tmp/session.pkl', main=None, byref=False):
     """pickle the current state of __main__ to a file"""
+    from .settings import settings
+    protocol = settings['protocol']
+    if main is None: main = _main_module
     f = open(filename, 'wb')
     try:
-        pickler = Pickler(f, 2)
-        pickler._main_module = main_module
-        _byref = pickler._byref
-        pickler._byref = False  # disable pickling by name reference
-        pickler._session = True # is best indicator of when pickling a session
-        pickler.dump(main_module)
-        pickler._session = False
-        pickler._byref = _byref
+        if byref:
+            main = _stash_modules(main)
+        pickler = Pickler(f, protocol)
+        pickler._main = main
+        pickler._byref = False   # disable pickling by name reference
+        pickler._recurse = False # disable pickling recursion for globals
+        pickler._session = True  # is best indicator of when pickling a session
+        pickler.dump(main)
     finally:
         f.close()
     return
 
-def load_session(filename='/tmp/session.pkl', main_module=_main_module):
+def load_session(filename='/tmp/session.pkl', main=None):
     """update the __main__ module with the state from the session file"""
+    if main is None: main = _main_module
     f = open(filename, 'rb')
     try:
         unpickler = Unpickler(f)
-        unpickler._main_module = main_module
+        unpickler._main = main
         unpickler._session = True
         module = unpickler.load()
         unpickler._session = False
-        main_module.__dict__.update(module.__dict__)
+        main.__dict__.update(module.__dict__)
+        _restore_modules(main)
     finally:
         f.close()
     return
 
 ### End: Pickle the Interpreter
 
+class MetaCatchingDict(dict):
+    def get(self, key, default=None):
+        try:
+            return self[key]
+        except KeyError:
+            return default
+
+    def __missing__(self, key):
+        if issubclass(key, type):
+            return save_type
+        else:
+            raise KeyError()
+
+
 ### Extend the Picklers
 class Pickler(StockPickler):
     """python's Pickler extended to interpreter sessions"""
-    dispatch = StockPickler.dispatch.copy()
-    _main_module = None
+    dispatch = MetaCatchingDict(StockPickler.dispatch.copy())
+    _main = None
     _session = False
-    _byref = False
-    pass
+    from .settings import settings
+    _byref = settings['byref']
+    _strictio = False
+    _fmode = settings['fmode']
+    _recurse = settings['recurse']
 
-    def __init__(self, *args, **kwargs):
-        StockPickler.__init__(self, *args, **kwargs)
-        self._main_module = _main_module
+    def __init__(self, *args, **kwds):
+        _byref = kwds.pop('byref', Pickler._byref)
+       #_strictio = kwds.pop('strictio', Pickler._strictio)
+        _fmode = kwds.pop('fmode', Pickler._fmode)
+        _recurse = kwds.pop('recurse', Pickler._recurse)
+        StockPickler.__init__(self, *args, **kwds)
+        self._main = _main_module
+        self._diff_cache = {}
+        self._byref = _byref
+       #self._strictio = _strictio
+        self._fmode = _fmode
+        self._recurse = _recurse
+    pass
 
 class Unpickler(StockUnpickler):
     """python's Unpickler extended to interpreter sessions and more types"""
-    _main_module = None
+    _main = None
     _session = False
 
     def find_class(self, module, name):
         if (module, name) == ('__builtin__', '__main__'):
-            return self._main_module.__dict__ #XXX: above set w/save_module_dict
+            return self._main.__dict__ #XXX: above set w/save_module_dict
         return StockUnpickler.find_class(self, module, name)
-    pass
 
-    def __init__(self, *args, **kwargs):
-        StockUnpickler.__init__(self, *args, **kwargs)
-        self._main_module = _main_module
+    def __init__(self, *args, **kwds):
+        StockUnpickler.__init__(self, *args, **kwds)
+        self._main = _main_module
+    pass
 
 '''
 def dispatch_table():
@@ -236,6 +363,8 @@ def dispatch_table():
     return Pickler.dispatch
 '''
 
+pickle_dispatch_copy = StockPickler.dispatch.copy()
+
 def pickle(t, func):
     """expose dispatch table for user-created extensions"""
     Pickler.dispatch[t] = func
@@ -247,6 +376,29 @@ def register(t):
         return func
     return proxy
 
+def _revert_extension():
+    for type, func in list(StockPickler.dispatch.items()):
+        if func.__module__ == __name__:
+            del StockPickler.dispatch[type]
+            if type in pickle_dispatch_copy:
+                StockPickler.dispatch[type] = pickle_dispatch_copy[type]
+
+def use_diff(on=True):
+    """
+    reduces size of pickles by only including object which have changed.
+    Decreases pickle size but increases CPU time needed.
+    Also helps avoid some unpicklable objects.
+    MUST be called at start of script, otherwise changes will not be recorded.
+    """
+    global _use_diff, diff
+    _use_diff = on
+    if _use_diff and diff is None:
+        try:
+            from . import diff as d
+        except:
+            import diff as d
+        diff = d
+
 def _create_typemap():
     import types
     if PY3:
@@ -310,6 +462,10 @@ def _create_function(fcode, fglobals, fname=None, fdefaults=None, \
     return func
 
 def _create_ftype(ftypeobj, func, args, kwds):
+    if kwds is None:
+        kwds = {}
+    if args is None:
+        args = ()
     return ftypeobj(func, *args, **kwds)
 
 def _create_lock(locked, *args):
@@ -320,27 +476,92 @@ def _create_lock(locked, *args):
             raise UnpicklingError("Cannot acquire lock")
     return lock
 
-def _create_filehandle(name, mode, position, closed, open=open): # buffering=0
+# thanks to matsjoyce for adding all the different file modes
+def _create_filehandle(name, mode, position, closed, open, strictio, fmode, fdata): # buffering=0
     # only pickles the handle, not the file contents... good? or StringIO(data)?
     # (for file contents see: http://effbot.org/librarybook/copy-reg.htm)
     # NOTE: handle special cases first (are there more special cases?)
     names = {'<stdin>':sys.__stdin__, '<stdout>':sys.__stdout__,
              '<stderr>':sys.__stderr__} #XXX: better fileno=(0,1,2) ?
-    if name in list(names.keys()): f = names[name] #XXX: safer "f=sys.stdin"
-    elif name == '<tmpfile>': import os; f = os.tmpfile()
-    elif name == '<fdopen>': import tempfile; f = tempfile.TemporaryFile(mode)
+    if name in list(names.keys()):
+        f = names[name] #XXX: safer "f=sys.stdin"
+    elif name == '<tmpfile>':
+        f = os.tmpfile()
+    elif name == '<fdopen>':
+        import tempfile
+        f = tempfile.TemporaryFile(mode)
     else:
-        try: # try to open the file by name   # NOTE: has different fileno
-            f = open(name, mode)#FIXME: missing: *buffering*, encoding,softspace
-        except IOError: 
+        # treat x mode as w mode
+        if "x" in mode and sys.hexversion < 0x03030000:
+            raise ValueError("invalid mode: '%s'" % mode)
+
+        if not os.path.exists(name):
+            if strictio:
+                raise FileNotFoundError("[Errno 2] No such file or directory: '%s'" % name)
+            elif "r" in mode and fmode != FILE_FMODE:
+                name = '<fdopen>' # or os.devnull?
+            current_size = 0 # or maintain position?
+        else:
+            current_size = os.path.getsize(name)
+
+        if position > current_size:
+            if strictio:
+                raise ValueError("invalid buffer size")
+            elif fmode == CONTENTS_FMODE:
+                position = current_size
+        # try to open the file by name
+        # NOTE: has different fileno
+        try:
+            #FIXME: missing: *buffering*, encoding, softspace
+            if fmode == FILE_FMODE:
+                f = open(name, mode if "w" in mode else "w")
+                f.write(fdata)
+                if "w" not in mode:
+                    f.close()
+                    f = open(name, mode)
+            elif name == '<fdopen>': # file did not exist
+                import tempfile
+                f = tempfile.TemporaryFile(mode)
+            elif fmode == CONTENTS_FMODE \
+               and ("w" in mode or "x" in mode):
+                # stop truncation when opening
+                flags = os.O_CREAT
+                if "+" in mode:
+                    flags |= os.O_RDWR
+                else:
+                    flags |= os.O_WRONLY
+                f = os.fdopen(os.open(name, flags), mode)
+                # set name to the correct value
+                if PY3:
+                    r = getattr(f, "buffer", f)
+                    r = getattr(r, "raw", r)
+                    r.name = name
+                else:
+                    class FILE(ctypes.Structure):
+                        _fields_ = [("refcount", ctypes.c_long),
+                                    ("type_obj", ctypes.py_object),
+                                    ("file_pointer", ctypes.c_voidp),
+                                    ("name", ctypes.py_object)]
+
+                    class PyObject(ctypes.Structure):
+                        _fields_ = [
+                            ("ob_refcnt", ctypes.c_int),
+                            ("ob_type", ctypes.py_object)
+                            ]
+                    if not HAS_CTYPES:
+                        raise ImportError("No module named 'ctypes'")
+                    ctypes.cast(id(f), ctypes.POINTER(FILE)).contents.name = name
+                    ctypes.cast(id(name), ctypes.POINTER(PyObject)).contents.ob_refcnt += 1
+                assert f.name == name
+            else:
+                f = open(name, mode)
+        except (IOError, FileNotFoundError):
             err = sys.exc_info()[1]
-            try: # failing, then use /dev/null #XXX: better to just fail here?
-                import os; f = open(os.devnull, mode)
-            except IOError:
-                raise UnpicklingError(err)
-                #XXX: python default is closed '<uninitialized file>' file/mode
-    if closed: f.close()
-    elif position >= 0: f.seek(position)
+            raise UnpicklingError(err)
+    if closed:
+        f.close()
+    elif position >= 0 and fmode != HANDLE_FMODE:
+        f.seek(position)
     return f
 
 def _create_stringi(value, position, closed):
@@ -378,12 +599,16 @@ class _attrgetter_helper(object):
             attrs[index] = ".".join([attrs[index], attr])
         return type(self)(attrs, index)
 
-if HAS_CTYPES:
-    ctypes.pythonapi.PyCell_New.restype = ctypes.py_object
-    ctypes.pythonapi.PyCell_New.argtypes = [ctypes.py_object]
-    # thanks to Paul Kienzle for cleaning the ctypes CellType logic
-    def _create_cell(contents):
-        return ctypes.pythonapi.PyCell_New(contents)
+if HAS_CTYPES and IS_PYPY:
+    try: # if using `pypi`, pythonapi is not found
+        ctypes.pythonapi.PyCell_New.restype = ctypes.py_object
+        ctypes.pythonapi.PyCell_New.argtypes = [ctypes.py_object]
+        # thanks to Paul Kienzle for cleaning the ctypes CellType logic
+        def _create_cell(contents):
+            return ctypes.pythonapi.PyCell_New(contents)
+
+    except AttributeError:
+        IS_PYPY = False
 
 def _create_weakref(obj, *args):
     from weakref import ref
@@ -409,6 +634,14 @@ def _create_weakproxy(obj, callable=False, *args):
 def _eval_repr(repr_str):
     return eval(repr_str)
 
+def _create_array(f, args, state, npdict=None):
+   #array = numpy.core.multiarray._reconstruct(*args)
+    array = f(*args)
+    array.__setstate__(state)
+    if npdict is not None: # we also have saved state in __dict__
+        array.__dict__.update(npdict)
+    return array
+
 def _getattr(objclass, name, repr_str):
     # hack to grab the reference directly
     try: #XXX: works only for __builtin__ ?
@@ -445,7 +678,7 @@ def _import_module(import_name, safe=False):
         raise
 
 def _locate_function(obj, session=False):
-    if obj.__module__ == '__main__': # and session:
+    if obj.__module__ in ['__main__', None]: # and session:
         return False
     found = _import_module(obj.__module__ + '.' + obj.__name__, safe=True)
     return found is obj
@@ -460,14 +693,20 @@ def save_code(pickler, obj):
 def save_function(pickler, obj):
     if not _locate_function(obj): #, pickler._session):
         log.info("F1: %s" % obj)
+        if getattr(pickler, '_recurse', False):
+            # recurse to get all globals referred to by obj
+            from .detect import globalvars
+            globs = globalvars(obj, recurse=True)
+        else:
+            globs = obj.__globals__ if PY3 else obj.func_globals
         if PY3:
-            pickler.save_reduce(_create_function, (obj.__code__, 
-                                obj.__globals__, obj.__name__,
+            pickler.save_reduce(_create_function, (obj.__code__,
+                                globs, obj.__name__,
                                 obj.__defaults__, obj.__closure__,
                                 obj.__dict__), obj=obj)
         else:
             pickler.save_reduce(_create_function, (obj.func_code,
-                                obj.func_globals, obj.func_name,
+                                globs, obj.func_name,
                                 obj.func_defaults, obj.func_closure,
                                 obj.__dict__), obj=obj)
     else:
@@ -477,7 +716,7 @@ def save_function(pickler, obj):
 
 @register(dict)
 def save_module_dict(pickler, obj):
-    if is_dill(pickler) and obj == pickler._main_module.__dict__ and not pickler._session:
+    if is_dill(pickler) and obj == pickler._main.__dict__ and not pickler._session:
         log.info("D1: <dict%s" % str(obj.__repr__).split('dict')[-1]) # obj
         if PY3:
             pickler.write(bytes('c__builtin__\n__main__\n', 'UTF-8'))
@@ -505,7 +744,7 @@ def save_module_dict(pickler, obj):
     return
 
 @register(ClassType)
-def save_classobj(pickler, obj):
+def save_classobj(pickler, obj): #FIXME: enable pickler._byref
     if obj.__module__ == '__main__': #XXX: use _main_module.__name__ everywhere?
         log.info("C1: %s" % obj)
         pickler.save_reduce(ClassType, (obj.__name__, obj.__bases__,
@@ -539,15 +778,8 @@ def save_attrgetter(pickler, obj):
     pickler.save_reduce(type(obj), tuple(attrs), obj=obj)
     return
 
-# __getstate__ explicitly added to raise TypeError when pickling:
-# http://www.gossamer-threads.com/lists/python/bugs/871199
-@register(FileType) #XXX: in 3.x has buffer=0, needs different _create?
-@register(BufferedRandomType)
-@register(BufferedReaderType)
-@register(BufferedWriterType)
-@register(TextWrapperType)
-def save_file(pickler, obj):
-    log.info("Fi: %s" % obj)
+def _save_file(pickler, obj, open_):
+    obj.flush()
     if obj.closed:
         position = None
     else:
@@ -555,24 +787,41 @@ def save_file(pickler, obj):
             position = -1
         else:
             position = obj.tell()
-    pickler.save_reduce(_create_filehandle, (obj.name, obj.mode, position, \
-                                             obj.closed), obj=obj)
+    if is_dill(pickler) and pickler._fmode == FILE_FMODE:
+        f = open_(obj.name, "r")
+        fdata = f.read()
+        f.close()
+    else:
+        fdata = ""
+    if is_dill(pickler):
+        strictio = pickler._strictio
+        fmode = pickler._fmode
+    else:
+        strictio = False
+        fmode = 0 # HANDLE_FMODE
+    pickler.save_reduce(_create_filehandle, (obj.name, obj.mode, position,
+                                             obj.closed, open_, strictio,
+                                             fmode, fdata), obj=obj)
     return
 
-if PyTextWrapperType: #XXX: are stdout, stderr or stdin ever _pyio files?
+
+@register(FileType) #XXX: in 3.x has buffer=0, needs different _create?
+@register(BufferedRandomType)
+@register(BufferedReaderType)
+@register(BufferedWriterType)
+@register(TextWrapperType)
+def save_file(pickler, obj):
+    log.info("Fi: %s" % obj)
+    return _save_file(pickler, obj, open)
+
+if PyTextWrapperType:
     @register(PyBufferedRandomType)
     @register(PyBufferedReaderType)
     @register(PyBufferedWriterType)
     @register(PyTextWrapperType)
     def save_file(pickler, obj):
         log.info("Fi: %s" % obj)
-        if obj.closed:
-            position = None
-        else:
-            position = obj.tell()
-        pickler.save_reduce(_create_filehandle, (obj.name, obj.mode, position, \
-                                                 obj.closed, _open), obj=obj)
-        return
+        return _save_file(pickler, obj, _open)
 
 # The following two functions are based on 'saveCStringIoInput'
 # and 'saveCStringIoOutput' from spickle
@@ -634,7 +883,7 @@ def save_instancemethod0(pickler, obj):# example: cStringIO.StringI
                                          obj.im_class), obj=obj)
     return
 
-if hex(sys.hexversion) >= '0x20500f0':
+if sys.hexversion >= 0x20500f0:
     @register(MemberDescriptorType)
     @register(GetSetDescriptorType)
     @register(MethodDescriptorType)
@@ -659,7 +908,7 @@ else:
                                        obj.__repr__()), obj=obj)
         return
 
-if HAS_CTYPES:
+if HAS_CTYPES and IS_PYPY:
     @register(CellType)
     def save_cell(pickler, obj):
         log.info("Ce: %s" % obj)
@@ -669,18 +918,19 @@ if HAS_CTYPES:
 # The following function is based on 'saveDictProxy' from spickle
 # Copyright (c) 2011 by science+computing ag
 # License: http://www.apache.org/licenses/LICENSE-2.0
-@register(DictProxyType)
-def save_dictproxy(pickler, obj):
-    log.info("Dp: %s" % obj)
-    attr = obj.get('__dict__')
-   #pickler.save_reduce(_create_dictproxy, (attr,'nested'), obj=obj)
-    if type(attr) == GetSetDescriptorType and attr.__name__ == "__dict__" \
-    and getattr(attr.__objclass__, "__dict__", None) == obj:
-        pickler.save_reduce(getattr, (attr.__objclass__, "__dict__"), obj=obj)
-        return
-    # all bad below... so throw ReferenceError or TypeError
-    from weakref import ReferenceError
-    raise ReferenceError("%s does not reference a class __dict__" % obj)
+if IS_PYPY:
+    @register(DictProxyType)
+    def save_dictproxy(pickler, obj):
+        log.info("Dp: %s" % obj)
+        attr = obj.get('__dict__')
+       #pickler.save_reduce(_create_dictproxy, (attr,'nested'), obj=obj)
+        if type(attr) == GetSetDescriptorType and attr.__name__ == "__dict__" \
+        and getattr(attr.__objclass__, "__dict__", None) == obj:
+            pickler.save_reduce(getattr, (attr.__objclass__,"__dict__"),obj=obj)
+            return
+        # all bad below... so throw ReferenceError or TypeError
+        from weakref import ReferenceError
+        raise ReferenceError("%s does not reference a class __dict__" % obj)
 
 @register(SliceType)
 def save_slice(pickler, obj):
@@ -697,7 +947,7 @@ def save_singleton(pickler, obj):
     return
 
 # thanks to Paul Kienzle for pointing out ufuncs didn't pickle
-if NumpyUfuncType:
+if NumpyArrayType:
     @register(NumpyUfuncType)
     def save_numpy_ufunc(pickler, obj):
         log.info("Nu: %s" % obj)
@@ -740,7 +990,7 @@ def _locate_object(address, module=None):
 @register(ReferenceType)
 def save_weakref(pickler, obj):
     refobj = obj()
-    log.info("Rf: %s" % obj)
+    log.info("R1: %s" % obj)
    #refobj = ctypes.pythonapi.PyWeakref_GetObject(obj) # dead returns "None"
     pickler.save_reduce(_create_weakref, (refobj,), obj=obj)
     return
@@ -749,8 +999,8 @@ def save_weakref(pickler, obj):
 @register(CallableProxyType)
 def save_weakproxy(pickler, obj):
     refobj = _locate_object(_proxy_helper(obj))
-    try: log.info("Rf: %s" % obj)
-    except ReferenceError: log.info("Rf: %s" % sys.exc_info()[1])
+    try: log.info("R2: %s" % obj)
+    except ReferenceError: log.info("R3: %s" % sys.exc_info()[1])
    #callable = bool(getattr(refobj, '__call__', None))
     if type(obj) is CallableProxyType: callable = True
     else: callable = False
@@ -759,21 +1009,43 @@ def save_weakproxy(pickler, obj):
 
 @register(ModuleType)
 def save_module(pickler, obj):
-    # if a module file name starts with this, it should be a standard module,
-    # so should be pickled as a reference
-    prefix = sys.base_prefix if PY3 else sys.prefix
-    std_mod = getattr(obj, "__file__", prefix).startswith(prefix)
-    if obj.__name__ not in ("builtins", "dill") \
-       and not std_mod or is_dill(pickler) and obj is pickler._main_module:
-        log.info("M1: %s" % obj)
-        _main_dict = obj.__dict__.copy() #XXX: better no copy? option to copy?
-        [_main_dict.pop(item, None) for item in singletontypes
-         + ["__builtins__", "__loader__"]]
-        pickler.save_reduce(_import_module, (obj.__name__,), obj=obj,
-                            state=_main_dict)
-    else:
+    if False: #_use_diff:
+        if obj.__name__ != "dill":
+            try:
+                changed = diff.whats_changed(obj, seen=pickler._diff_cache)[0]
+            except RuntimeError:  # not memorised module, probably part of dill
+                pass
+            else:
+                log.info("M1: %s with diff" % obj)
+                log.info("Diff: %s", changed.keys())
+                pickler.save_reduce(_import_module, (obj.__name__,), obj=obj,
+                                    state=changed)
+                return
+
         log.info("M2: %s" % obj)
         pickler.save_reduce(_import_module, (obj.__name__,), obj=obj)
+    else:
+        # if a module file name starts with prefx, it should be a builtin
+        # module, so should be pickled as a reference
+        if hasattr(obj, "__file__"):
+            names = ["base_prefix", "base_exec_prefix", "exec_prefix",
+                     "prefix", "real_prefix"]
+            builtin_mod = any([obj.__file__.startswith(getattr(sys, name))
+                           for name in names if hasattr(sys, name)])
+        else:
+            builtin_mod = True
+        if obj.__name__ not in ("builtins", "dill") \
+           and not builtin_mod or is_dill(pickler) and obj is pickler._main:
+            log.info("M1: %s" % obj)
+            _main_dict = obj.__dict__.copy() #XXX: better no copy? option to copy?
+            [_main_dict.pop(item, None) for item in singletontypes
+                + ["__builtins__", "__loader__"]]
+            pickler.save_reduce(_import_module, (obj.__name__,), obj=obj,
+                                state=_main_dict)
+        else:
+            log.info("M2: %s" % obj)
+            pickler.save_reduce(_import_module, (obj.__name__,), obj=obj)
+        return
     return
 
 @register(TypeType)
@@ -789,7 +1061,7 @@ def save_type(pickler, obj):
             StockPickler.save_global(pickler, obj)
             return
         except AttributeError: pass
-        if type(obj) == type:
+        if issubclass(type(obj), type):
         #   try: # used when pickling the class as code (or the interpreter)
             if is_dill(pickler) and not pickler._byref:
                 # thanks to Tom Stepleton pointing out pickler._session unneeded
@@ -816,6 +1088,12 @@ def save_type(pickler, obj):
         StockPickler.save_global(pickler, obj)
     return
 
+@register(property)
+def save_property(pickler, obj):
+    log.info("Pr: %s" % obj)
+    pickler.save_reduce(property, (obj.fget, obj.fset, obj.fdel, obj.__doc__),
+                        obj=obj)
+
 # quick sanity checking
 def pickles(obj,exact=False,safe=False,**kwds):
     """quick check if object pickles with dill"""
@@ -824,9 +1102,17 @@ def pickles(obj,exact=False,safe=False,**kwds):
         exceptions = (TypeError, AssertionError, PicklingError, UnpicklingError)
     try:
         pik = copy(obj, **kwds)
-        if exact:
-            return pik == obj
-        return type(pik) == type(obj)
+        try:
+            result = bool(pik.all() == obj.all())
+        except AttributeError:
+            result = pik == obj
+        if result: return True
+        if not exact:
+            result = type(pik) == type(obj)
+            if result: return result
+            # class instances might have been dumped with byref=False
+            return repr(type(pik)) == repr(type(obj)) #XXX: InstanceType?
+        return False
     except exceptions:
         return False
 
@@ -834,7 +1120,7 @@ def pickles(obj,exact=False,safe=False,**kwds):
 def is_dill(pickler):
     "check the dill-ness of your pickler"
     return 'dill' in pickler.__module__
-   #return hasattr(pickler,'_main_module')
+   #return hasattr(pickler,'_main')
 
 def _extend():
     """extend pickle with all of dill's registered types"""
@@ -847,4 +1133,6 @@ def _extend():
         else: pass
     return
 
+del diff, _use_diff, use_diff
+
 # EOF
